package vishnu.cluster;

public class Hierarchical
{
    public float[] minrow;
    public int[] colind;
    public TriangleIndex triangle;
    
    public Hierarchical(int matrix_rows){
	// create Triangle with "limit" arg = maximum #clusters
	// an object to help with indexing the cells 
	// in a triangular matrix
	triangle = new TriangleIndex(matrix_rows);
    }
    
    
    // set each minrow[i] to the value of the minimum element of that row
    void set_minrow (SimMatrix m, int row)
    {
	// triangular matrix without diagonal, hence zeroth row is all zero
        if(row == 0) return;
	
	int i;
	int index_start, index_end;
        
        // first column
	index_start = triangle.index (row, 0);

	// last column (row(!)-1 because as many rows as columns)
	index_end   = triangle.index (row, row - 1); 
	
	minrow[row] = m.matrix[index_start];
	colind[row] = index_start;
	
	for (i = index_start + 1; i <= index_end; i++){
	    if (m.matrix[i] < minrow[row]){
		minrow[row] = m.matrix[i];
		colind[row] = i;
	    }
	}
    }
    
    
    // return the ROW in which the minimum disimilarity resides
    int get_minrow (int last_row){
	int i;
	int row;
	float value;
	
	row   = 1;
	value = minrow[1];
	
	for (i = 2; i <= last_row; i++){
	    if (minrow[i] < value){
		row = i;
		value = minrow[i];
	    }
    	}
	
	return row;
    }

    
    // num_of_rows is "limit" = max docs to be clustered in hierarchical clustering
    Cluster clustering (SimMatrix sim_matrix, int num_of_rows, double[] heights, LinkFunc linkage_func)
    {
	Cluster child_1;
	Cluster child_2;
	Cluster parent=null;
	ClusterArray cluster_ary;
	
        int min_row, index, row, col, i;
	double sim;
	
        int last_row_index, sim_matrix_size;
	
        sim_matrix_size = num_of_rows *(num_of_rows - 1) / 2;
	
	// Place each of the documents in a cluster of each own 
	// there are vector.length docs but only num_of_rows = limit places in the array
	// may be two-stage clustering process, i.e. cluster subset of docs and then add the rest
	cluster_ary = new ClusterArray (num_of_rows);
	last_row_index = num_of_rows - 1;
	
	// minimum distance for each row of the sim_matrix
	minrow = new float[num_of_rows];
	
	// column index for minimum distance entry
	colind = new int[num_of_rows];
	
	for (i = 1; i < num_of_rows; i++) set_minrow(sim_matrix, i);
	
	// do all the hierarchical clustering by running over all rows, updating the matrix 
	// as a new cluster is formed each time
	for (i = 0; i < num_of_rows - 1; i++){

	    // find the row with the lowest disimilarity
	    min_row = get_minrow(last_row_index);
	    sim = minrow[min_row];
	    
	    System.out.println("Min sim: " + sim);
	    
	    // the column with the lowest distance in that row
	    index  = colind[min_row];
	    
	    // array of heights, DESCENDING disim, i.e. sim goes last
	    heights[num_of_rows - 2 - i] = sim; 
	    
	    row = triangle.i_ind[index];
	    col = triangle.j_ind[index];
	    
	    // merges rows and cols to create a smaller sim_matrix
	    merge (row, col, sim_matrix, last_row_index, linkage_func);
	    System.out.println("Merging: " + row + " and " + col);
	    
	    sim_matrix_size -= last_row_index;
	    
	    child_1 = (Cluster)cluster_ary.contents.elementAt(row);
	    child_2 = (Cluster)cluster_ary.contents.elementAt(col);
	    
	    // parents get id = -1 to indicate that they are not representing single docs
	    // all leaves have an id (1 - num_rows) that points to the doc they contain
	    int sum = child_1.items + child_2.items;
	    System.out.println("Create new parent with " + child_1.items + " + " + child_2.items + " leaves");
	    parent = new Cluster (-1, null, child_1, child_2, sim, child_1.items + child_2.items);
	    
	    child_1.parent = parent;
	    child_2.parent = parent;
	    
	    cluster_ary.contents.setElementAt(cluster_ary.contents.elementAt(last_row_index),row);
	    cluster_ary.contents.setElementAt(parent,col);
	    
	    last_row_index--;
	}
	
	return parent;
    }
    
    // form a new cluster from the two most similar documents
    void merge (int i, int j, SimMatrix m, int n, LinkFunc linkage_func)
    {
	
	/* i and j are the matrix elements with maximum similarity         */
	/* matrix is the triangle similarity matrix with last row index n  */
	/* linkage_func is a pointer to the linkage function               */
	
	/* i > j */
	
	int c, k, index1, index2;
	
        // go over all rows
	for (k = 0; k <= n; k++){
	    
    	    // if k < j the affected elements are on row j
	    if (k < j){
		index1 = triangle.index (k, j);
		index2 = triangle.index (k, i);
		m.matrix[index1] = linkage_func.link(m.matrix[index1], m.matrix[index2]);
	    }else if (k > j && k < i){
		/* if j < k < i then only one element is updated per row */
		index1 = triangle.index (k, j);
		index2 = triangle.index (k, i);
		m.matrix[index1] = linkage_func.link(m.matrix[index1], m.matrix[index2]);
		
		if (index1 == colind[k]){
		    /* the previous minimum of the row has been changed so recaculate minimum of whole row */
		    set_minrow(m, k);
		}else{
		    /* if the changed element is smaller than the current minimum update minrow and colind */
		    if (m.matrix[index1] < minrow[k]){
			minrow[k] = m.matrix[index1];
			colind[k] = index1;
		    }
		}
	    }else if (k > i && k < n){
		index1 = triangle.index (k, j);
		index2 = triangle.index (k, i);
		m.matrix[index1] = linkage_func.link(m.matrix[index1], m.matrix[index2]);
		//printf ("3 - [%d, %d] = link ([%d, %d], [%d, %d]) = %f\n", k, j, k, i, k, j, matrix[index2]);
		
		if (index1 == colind[k]){
		    /* the previous minimum of the row has been changed so recaculate minimum of whole row */
		    set_minrow(m, k);
		}else{
		    /* if the changed element is smaller than the current minimum update minrow and colind */
		    if (m.matrix[index1] < minrow[k]){
			minrow[k] = m.matrix[index1];
			colind[k] = index1;
		    }
		}
		
		index1 = triangle.index (k, i);
		index2 = triangle.index (k, n);
		m.matrix[index1] = m.matrix[index2];
		
		if (index1 == colind[k]){
		    /* the previous minimum of the row has been changed so recaculate minimum of whole row */
		    set_minrow(m, k);
		}else{
		    /* if the changed element is smaller than the current minimum update minrow and colind */
		    if (m.matrix[index1] < minrow[k]){
			minrow[k] = m.matrix[index1];
			colind[k] = index1;
		    }
		}
	    }else if(k == n  && i != n){
		index1 = triangle.index (k, j);
		index2 = triangle.index (k, i);
		m.matrix[index1] = linkage_func.link(m.matrix[index1], m.matrix[index2]);
		
		if (index1 == colind[k]){
		    /* the previous minimum of the row has been changed so recaculate minimum of whole row */
		    set_minrow(m, k);
		}else{
		    /* if the changed element is smaller than the current minimum update minrow and colind */
		    if (m.matrix[index1] < minrow[k]){
			minrow[k] = m.matrix[index1];
			colind[k] = index1;
		    }
		}
		
		for (c = 0; c < i; c++){
		    index1 = triangle.index (c, i);
		    index2 = triangle.index (c, n);
		    m.matrix[index1] = m.matrix[index2];
		}
		
	    }
    	}
	set_minrow(m, i);
	set_minrow(m, j);
    }
}

